#include <spdlog/spdlog.h>

#include "include/GoToController.hpp"
#include "include/Robot.hpp"
#include "include/ClosedLoopMotorController.hpp"
#include "include/OdometryController.hpp"
#include "include/mathUtils.hpp"

#define MS_IN_S 1000
#define US_IN_S (1000 * MS_IN_S)

// Source: http://faculty.salina.k-state.edu/tim/robot_prog/MobileBot/Steering/unicycle.html#calculating-wheel-velocities
// Source: https://www.roboticsbook.org/S52_diffdrive_actions.html#kinematics-in-code
void GoToController::diff_drive_inverse_kinematics(double v_m_s, double w_rad_s) {
  auto vr_rad_s = (2 * v_m_s - w_rad_s * ROBOT_ARBOR_LENGTH_M) / (2.0 * ROBOT_WHEEL_RADIUS_M) * ROBOT_ODOMETRY_CONTROLLER_RIGHT_MULT;
  auto vl_rad_s = (2 * v_m_s + w_rad_s * ROBOT_ARBOR_LENGTH_M) / (2.0 * ROBOT_WHEEL_RADIUS_M) * ROBOT_ODOMETRY_CONTROLLER_LEFT_MULT;

  ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_RIGHT_PORT, vr_rad_s);
  ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_LEFT_PORT, vl_rad_s);
}

void GoToController::go_to_point(double goal_x, double goal_y, double v_ms) {
  auto was_enabled = OdometryController::getInstance().is_enabled();
  OdometryController::getInstance().enable();
  auto start_pos = OdometryController::getInstance().get();

  auto sleep_duration = std::chrono::microseconds(US_IN_S / ROBOT_MOTOR_SPEED_CONTROL_RATE_HZ);

  auto abs_distance = fabs(sqrt(
      pow(goal_x - start_pos.get_x_position(), 2) + pow(goal_y - start_pos.get_y_position(), 2)
  ));
  while (abs_distance > 0.05) {
    auto current_pos = OdometryController::getInstance().get();
    auto alpha = mathUtils::wrap_angle_to_pi(
        (atan2(goal_x - current_pos.get_x_position(), goal_y - current_pos.get_y_position()) - (current_pos.get_angular_orientation() + M_PI_2))
    );

    auto vr_ms = v_ms * (cos(alpha) - ROBOT_GO_TO_GOAL_K * sin(alpha));
    auto vl_ms = v_ms * (cos(alpha) + ROBOT_GO_TO_GOAL_K * sin(alpha));
    auto vr_rad_s = vr_ms / ROBOT_WHEEL_RADIUS_M * ROBOT_ODOMETRY_CONTROLLER_RIGHT_MULT;
    auto vl_rad_s = vl_ms / ROBOT_WHEEL_RADIUS_M * ROBOT_ODOMETRY_CONTROLLER_LEFT_MULT;

    ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_RIGHT_PORT, vr_rad_s);
    ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_LEFT_PORT, vl_rad_s);

    abs_distance = fabs(sqrt(
        pow(goal_x - current_pos.get_x_position(), 2) + pow(goal_y - current_pos.get_y_position(), 2)
    ));
    spdlog::debug("{:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}",
                  current_pos.get_x_position(), current_pos.get_y_position(), current_pos.get_angular_orientation(),
                  goal_x - current_pos.get_x_position(), goal_y - current_pos.get_y_position(),
                  current_pos.get_angular_orientation() + M_PI_2, atan2(goal_x - current_pos.get_x_position(), goal_y - current_pos.get_y_position()), alpha,
                  vr_rad_s, vl_rad_s);

    std::this_thread::sleep_for(sleep_duration);
  }

  ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_RIGHT_PORT, 0);
  ClosedLoopMotorController::getInstance().set_speed(ROBOT_ODOMETRY_CONTROLLER_LEFT_PORT, 0);

  if (!was_enabled) {
    OdometryController::getInstance().disable();
  }
}

void GoToController::drive_distance(double distance_m, double v_m_s) {
  if (distance_m < 0 && v_m_s > 0) {
    distance_m = fabs(distance_m);
    v_m_s *= -1;
  }

  auto was_enabled = OdometryController::getInstance().is_enabled();
  OdometryController::getInstance().enable();
  auto start_pos = OdometryController::getInstance().get();

  auto sleep_duration = std::chrono::microseconds(US_IN_S / ROBOT_MOTOR_SPEED_CONTROL_RATE_HZ);

  auto abs_distance_to_target = distance_m;
  while (abs_distance_to_target > ROBOT_GO_TO_DISTANCE_END_M) {
    std::this_thread::sleep_for(sleep_duration);

    auto current_pos = OdometryController::getInstance().get();

    auto distance_driven = fabs(current_pos.get_x_position() - start_pos.get_x_position());
    abs_distance_to_target = fabs(distance_m - distance_driven);

    auto alpha = -mathUtils::wrap_angle_to_pi(current_pos.get_angular_orientation());

    auto vel_mult = fmax(0, fmin(1, abs_distance_to_target * ROBOT_GO_TO_DISTANCE_VELOCITY_P));
    diff_drive_inverse_kinematics(v_m_s * vel_mult, alpha * ROBOT_GO_TO_DISTANCE_ALPHA_P);

    spdlog::debug("{:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}",
                  current_pos.get_x_position(), current_pos.get_y_position(), current_pos.get_angular_orientation(),
                  distance_driven, abs_distance_to_target,
                  alpha, vel_mult
    );
  }

  ClosedLoopMotorController::getInstance().set_power(ROBOT_ODOMETRY_CONTROLLER_RIGHT_PORT, 0);
  ClosedLoopMotorController::getInstance().set_power(ROBOT_ODOMETRY_CONTROLLER_LEFT_PORT, 0);

  if (!was_enabled) {
    OdometryController::getInstance().disable();
  }
}

void GoToController::turn_degrees(double angle_deg, double v_rad_s) {
  if (angle_deg < 0 && v_rad_s > 0) {
    v_rad_s *= -1;
  }

  auto angle_rad = mathUtils::wrap_angle_to_pi(angle_deg * (M_PI / 180.0));

  auto was_enabled = OdometryController::getInstance().is_enabled();
  OdometryController::getInstance().enable();
  auto start_pos = OdometryController::getInstance().get();

  auto sleep_duration = std::chrono::microseconds(US_IN_S / ROBOT_MOTOR_SPEED_CONTROL_RATE_HZ);

  auto angle_to_target = angle_rad;
  while (fabs(angle_to_target) > ROBOT_GO_TO_ANGLE_END_RAD) {
    std::this_thread::sleep_for(sleep_duration);

    auto current_pos = OdometryController::getInstance().get();

    auto angle_turned = current_pos.get_angular_orientation() - start_pos.get_angular_orientation();
    angle_to_target = mathUtils::wrap_angle_to_pi(angle_rad - angle_turned);

    diff_drive_inverse_kinematics(0.0, v_rad_s);

    spdlog::debug("{:03.4f}; {:03.4f}; {:03.4f}; {:03.4f}; {:03.4f};",
                  current_pos.get_x_position(), current_pos.get_y_position(), current_pos.get_angular_orientation(),
                  angle_turned, angle_to_target
    );
  }

  ClosedLoopMotorController::getInstance().set_power(ROBOT_ODOMETRY_CONTROLLER_RIGHT_PORT, 0);
  ClosedLoopMotorController::getInstance().set_power(ROBOT_ODOMETRY_CONTROLLER_LEFT_PORT, 0);

  if (!was_enabled) {
    OdometryController::getInstance().disable();
  }
}
